import os

SARK_PLUGINS_ENV_NAME = "sarkPlugins"

SARK_DEFAULT_PATH = os.path.normpath(os.path.join(os.path.dirname(__file__), "../"))

def get_sark_dir(dirname, envname=None):
    if envname in os.environ:
        return os.environ[envname]

    return os.path.join(SARK_DEFAULT_PATH, dirname)


def get_plugins_dir():
    return get_sark_dir("plugins", SARK_PLUGINS_ENV_NAME)


def get_codecs_dir():
    from sark import encodings
    return os.path.dirname(encodings.__file__)

CODECS_DIR = get_codecs_dir()
PLUGINS_DIR = get_plugins_dir()


import idaapi
import idautils
import idc
from ..core import fix_addresses
from .xref import Xref
from .instruction import Instruction
from ..ui import updates_ui
from .base import get_selection, get_offset_name, demangle
from .. import data


class Comments(object):
    """IDA Line Comments

    Provides easy access to all types of comments for an IDA line.
    """

    def __init__(self, ea):
        self._ea = ea

    def __bool__(self):
        return any((self.regular, self.repeat, self.anterior, self.posterior,))

    @property
    def regular(self):
        """Regular Comment"""
        return idaapi.get_cmt(self._ea, 0)

    @regular.setter
    def regular(self, comment):
        idaapi.set_cmt(self._ea, comment, 0)

    @property
    def repeat(self):
        """Repeatable Comment"""
        return idaapi.get_cmt(self._ea, 1)

    @repeat.setter
    def repeat(self, comment):
        idaapi.set_cmt(self._ea, comment, 1)

    def _iter_extra_comments(self, start):
        end = idaapi.get_first_free_extra_cmtidx(self._ea, start)
        for idx in range(start, end):
            line = idaapi.get_extra_cmt(self._ea, idx)
            yield line or ''

    def _iter_anterior(self):
        return self._iter_extra_comments(idaapi.E_PREV)

    @property
    def anterior(self):
        """Anterior Comment"""
        return "\n".join(self._iter_anterior())

    @anterior.setter
    @updates_ui
    def anterior(self, comment):
        if not comment:
            idaapi.del_extra_cmt(self._ea, idaapi.E_PREV)
            return

        index = 0

        for index, line in enumerate(comment.splitlines()):
            idaapi.update_extra_cmt(self._ea, idaapi.E_PREV + index, line)

        idaapi.del_extra_cmt(self._ea, idaapi.E_PREV + (index + 1))

    def _iter_posterior(self):
        return self._iter_extra_comments(idaapi.E_NEXT)

    @property
    def posterior(self):
        """Posterior Comment"""
        return "\n".join(self._iter_posterior())

    @posterior.setter
    @updates_ui
    def posterior(self, comment):
        if not comment:
            idaapi.del_extra_cmt(self._ea, idaapi.E_NEXT)
            return

        index = 0

        for index, line in enumerate(comment.splitlines()):
            idaapi.update_extra_cmt(self._ea, idaapi.E_NEXT + index, line)

        idaapi.del_extra_cmt(self._ea, idaapi.E_NEXT + (index + 1))

    def __repr__(self):
        return ("Comments("
                "ea=0x{ea:08X},"
                " reqular={regular},"
                " repeat={repeat},"
                " anterior={anterior},"
                " posterior={posterior})").format(
            ea=self._ea,
            regular=repr(self.regular),
            repeat=repr(self.repeat),
            anterior=repr(self.anterior),
            posterior=repr(self.posterior))


class Line(object):
    """
    An IDA Line.

    This objects encapsulates many of IDA's line-handling APIs in an easy to use
    and object oriented way.
    """

    class UseCurrentAddress(object):
        """
        This is a filler object to replace `None` for the EA.
        In many cases, a programmer can accidentally initialize the
        `Line` object with `ea=None`, resulting in the current address.
        Usually, this is not the desired outcome. This object resolves this issue.
        """
        pass

    def __init__(self, ea=UseCurrentAddress, name=None):
        if name is not None and ea != self.UseCurrentAddress:
            raise ValueError(("Either supply a name or an address (ea). "
                              "Not both. (ea={!r}, name={!r})").format(ea, name))

        elif name is not None:
            ea = idc.get_name_ea_simple(name)

        elif ea == self.UseCurrentAddress:
            ea = idc.here()

        elif ea is None:
            raise ValueError("`None` is not a valid address. To use the current screen ea, "
                             "use `Line(ea=Line.UseCurrentAddress)` or supply no `ea`.")

        self._ea = idaapi.get_item_head(ea)
        self._comments = Comments(self._ea)

    @property
    def flags(self):
        """`FF_*` Flags. See `bytes.hpp`."""
        return idaapi.get_full_flags(self.ea)

    @property
    def is_code(self):
        """Is the line code."""
        return idaapi.is_code(self.flags)

    @property
    def is_data(self):
        """Is the line data."""
        return idaapi.is_data(self.flags)

    @property
    def is_unknown(self):
        """Is the line unknown."""
        return idaapi.is_unknown(self.flags)

    @property
    def is_tail(self):
        """Is the line a tail."""
        return idaapi.is_tail(self.flags)

    @property
    def is_string(self):
        """Is the line a string."""
        return data.is_string(self.ea)

    @property
    def comments(self):
        """Comments"""
        return self._comments

    @property
    def ea(self):
        """Line EA"""
        return self._ea

    start_ea = ea

    @property
    def end_ea(self):
        """End address of line (first byte after the line)"""
        return self.ea + self.size

    @property
    def disasm(self):
        """Line Disassembly"""
        return idc.GetDisasm(self.ea)

    @property
    def type(self):
        """return the type of the Line """
        properties = {self.is_code: "code",
                      self.is_data: "data",
                      self.is_string: "string",
                      self.is_tail: "tail",
                      self.is_unknown: "unknown"}
        for k, v in properties.items():
            if k: return v

    def __repr__(self):
        return "[{:08X}]    {}".format(self.ea, self.disasm)

    @property
    def xrefs_from(self):
        """Xrefs from this line.

        :return: Xrefs as `sark.code.xref.Xref` objects.
        """
        return list(map(Xref, idautils.XrefsFrom(self.ea)))

    @property
    def calls_from(self):
        return (xref for xref in self.xrefs_from if xref.type.is_call)

    @property
    def drefs_from(self):
        """Destination addresses of data references from this line."""
        return idautils.DataRefsFrom(self.ea)

    @property
    def crefs_from(self):
        """Destination addresses of code references from this line."""
        return idautils.CodeRefsFrom(self.ea, 1)

    @property
    def xrefs_to(self):
        """Xrefs to this line.

        Returns:
            Xrefs as `sark.code.xref.Xref` objects.
        """
        return list(map(Xref, idautils.XrefsTo(self.ea)))

    @property
    def drefs_to(self):
        """Source addresses of data references from this line."""
        return idautils.DataRefsTo(self.ea)

    @property
    def crefs_to(self):
        """Source addresses of data references to this line."""
        return idautils.CodeRefsTo(self.ea, 1)

    @property
    def size(self):
        """Size (in bytes) of the line."""
        return idaapi.get_item_size(self.ea)

    @property
    def name(self):
        """Name of the line (the label shown in IDA)."""
        return idaapi.get_ea_name(self.ea)

    @name.setter
    def name(self, value):
        idc.set_name(self.ea, value)

    @property
    def demangled(self):
        """Return the demangled name of the line. If none exists, return `.name`"""
        return demangle(self.name)

    @property
    def insn(self):
        """Instruction"""
        return Instruction(self.ea)

    @property
    def color(self):
        """Line color in IDA View"""
        color = idc.get_color(self.ea, idc.CIC_ITEM)
        if color == 0xFFFFFFFF:
            return None

        return color

    @color.setter
    @updates_ui
    def color(self, color):
        """Line Color in IDA View.

        Set color to `None` to clear the color.
        """
        if color is None:
            color = 0xFFFFFFFF

        idc.set_color(self.ea, idc.CIC_ITEM, color)

    @property
    def next(self):
        """The next line."""
        return Line(self.end_ea)

    @property
    def prev(self):
        """The previous line."""
        return Line(self.ea - 1)

    @property
    def has_name(self):
        """Does the current line have a non-trivial (non-dummy) name?"""
        return idaapi.has_name(self.flags)

    @property
    def offset_name(self):
        return get_offset_name(self.ea)

    @property
    def bytes(self):
        return idaapi.get_bytes(self.ea, self.size)

    def __eq__(self, other):
        if not isinstance(other, Line):
            return False

        return self.ea == other.ea

    def __ne__(self, other):
        return not self.__eq__(other)


def lines(start=None, end=None, reverse=False, selection=False):
    """Iterate lines in range.

    Args:
        start: Starting address, start of IDB if `None`.
        end: End address, end of IDB if `None`.
        reverse: Set to true to iterate in reverse order.
        selection: If set to True, replaces start and end with current selection.

    Returns:
        iterator of `Line` objects.
    """
    if selection:
        start, end = get_selection()

    else:
        start, end = fix_addresses(start, end)

    if not reverse:
        item = idaapi.get_item_head(start)
        while item < end:
            yield Line(item)
            item += idaapi.get_item_size(item)

    else:  # if reverse:
        item = idaapi.get_item_head(end - 1)
        while item >= start:
            yield Line(item)
            item = idaapi.get_item_head(item - 1)



import networkx

import idaapi
from .code import lines, functions
from .core import get_func, fix_addresses


class CodeBlock(idaapi.BasicBlock):
    def __init__(self, id_ea=None, bb=None, fc=None):
        if bb is None and fc is None:
            if id_ea is None:
                id_ea = idaapi.get_screen_ea()
            temp_codeblock = get_codeblock(id_ea)
            self.__dict__.update(temp_codeblock.__dict__)
        else:
            super(CodeBlock, self).__init__(id=id_ea, bb=bb, fc=fc)

    @property
    def lines(self):
        return lines(self.start_ea, self.end_ea)

    @property
    def next(self):
        return self.succs()

    @property
    def prev(self):
        return self.preds()

    def set_color(self, color=None):
        for line in self.lines:
            line.color = color

        if color is None:
            idaapi.clr_node_info(self._fc._q.bounds.start_ea, self.id, idaapi.NIF_BG_COLOR)

        else:
            node_info = idaapi.node_info_t()
            node_info.bg_color = color
            idaapi.set_node_info(self._fc._q.bounds.start_ea, self.id, node_info, idaapi.NIF_BG_COLOR)

    @property
    def color(self):
        node_info = idaapi.node_info_t()
        success = idaapi.get_node_info(node_info, self._fc._q.bounds.start_ea, self.id)

        if not success:
            return None

        if not node_info.valid_bg_color():
            return None

        return node_info.bg_color

    @color.setter
    def color(self, color):
        self.set_color(color)

    def __repr__(self):
        return "<CodeBlock(start_ea=0x{:08X}, end_ea=0x{:08X})>".format(self.start_ea, self.end_ea)

    def __eq__(self, other):
        return self.start_ea == other.start_ea


class FlowChart(idaapi.FlowChart):
    def __init__(self, f=None, bounds=None, flags=idaapi.FC_PREDS, ignore_external=False):
        if f is None and bounds is None:
            f = idaapi.get_screen_ea()
        if f is not None:
            f = get_func(f)
        if ignore_external:
            flags |= idaapi.FC_NOEXT

        super(FlowChart, self).__init__(f=f, bounds=bounds, flags=flags)

    def _getitem(self, index):
        return CodeBlock(index, self._q[index], self)


def get_flowchart(ea=None):
    if ea is None:
        ea = idaapi.get_screen_ea()
    func = idaapi.get_func(ea)
    flowchart_ = FlowChart(func)
    return flowchart_


def get_codeblock(ea=None):
    if ea is None:
        ea = idaapi.get_screen_ea()
    flowchart_ = get_flowchart(ea)
    for code_block in flowchart_:
        if code_block.start_ea == ea: # External blocks can be zero-sized.
            return code_block
        if code_block.start_ea <= ea < code_block.end_ea:
            return code_block


def get_block_start(ea):
    """Get the start address of an IDA Graph block."""
    return get_codeblock(ea).start_ea


def get_nx_graph(ea, ignore_external=False):
    """Convert an IDA flowchart to a NetworkX graph."""
    nx_graph = networkx.DiGraph()
    func = idaapi.get_func(ea)
    flowchart = FlowChart(func, ignore_external=ignore_external)
    for block in flowchart:
        # Make sure all nodes are added (including edge-less nodes)
        nx_graph.add_node(block.start_ea)

        for pred in block.preds():
            nx_graph.add_edge(pred.start_ea, block.start_ea)
        for succ in block.succs():
            nx_graph.add_edge(block.start_ea, succ.start_ea)

    return nx_graph


def codeblocks(start=None, end=None, full=True):
    """Get all `CodeBlock`s in a given range.

    Args:
        start - start address of the range. If `None` uses IDB start.
        end - end address of the range. If `None` uses IDB end.
        full - `True` is required to change node info (e.g. color). `False` causes faster iteration.
    """
    if full:
        for function in functions(start, end):
            fc = FlowChart(f=function.func_t)
            for block in fc:
                yield block

    else:
        start, end = fix_addresses(start, end)

        for code_block in FlowChart(bounds=(start, end)):
            yield code_block



import collections
import itertools
import networkx as nx
from .code.function import functions, Function
from contextlib import suppress
from . import exceptions


def lowest_common_ancestors(G, targets):
    common_ancestors = None
    all_ancestors = set()
    for target in targets:
        parents = set()
        q = collections.deque()
        q.append(target)

        while q:
            n = q.popleft()
            if n in parents:
                continue
            for p in G.predecessors(n):
                q.append(p)
            parents.add(n)

        all_ancestors.update(parents)

        if common_ancestors is None:
            common_ancestors = parents
        else:
            common_ancestors &= parents

    lowest_common = set()
    if common_ancestors is not None:
        for p in common_ancestors:
            if any(child not in common_ancestors and child in all_ancestors for child in G.successors(p)):
                lowest_common.add(p)

    return lowest_common


def _try_get_function_start(ea):
    with suppress(exceptions.SarkNoFunction):
        return Function(ea).start_ea

    return ea


def get_idb_graph():
    """Export IDB to a NetworkX graph.

    Use xrefs to and from functions to build a DiGraph containing all
    the functions in the IDB and all the links between them.
    The graph can later be used to perform analysis on the IDB.

    :return: nx.DiGraph()
    """
    digraph = nx.DiGraph()

    for function in functions():
        for xref in itertools.chain(function.xrefs_from, function.xrefs_to):
            frm = _try_get_function_start(xref.frm)
            to = _try_get_function_start(xref.to)

            digraph.add_edge(frm, to)

    return digraph


def get_lca_graph(G, targets, lca_sources=None):
    if lca_sources is None:
        lca_sources = lowest_common_ancestors(G, targets)

    lca_graph = nx.DiGraph()

    for source in lca_sources:
        for target in targets:
            path = nx.shortest_path(G, source, target)

            for frm, to in zip(path[:-1], path[1:]):
                lca_graph.add_edge(frm, to)

    return lca_graph


from collections import defaultdict


import idaapi

from .. import exceptions

class Switch(object):
    """IDA Switch

    Access IDA switch data with ease.

    Usage:

        >>> my_switch = Switch(switch_jump_address)
        >>> for case, target in my_switch:
        ...     print("{} -> 0x{:08X}".format(case, target))

    """
    def __init__(self, ea):
        """Initialize a switch parser.

        Args:
            ea: An address of a switch jump instruction.
        """
        self._ea = ea

        results = self._calc_cases()

        self._map = self._build_map(results)

        self._reverse_map = self._build_reverse(self._map)

    def _build_reverse(self, switch_map):
        reverse_map = defaultdict(list)
        for case, target in switch_map.items():
            reverse_map[target].append(case)
        return reverse_map

    def _calc_cases(self):
        si = idaapi.get_switch_info(self._ea)
        results = idaapi.calc_switch_cases(self._ea, si)
        if not results:
            raise exceptions.SarkNotASwitch("Seems like 0x{:08X} is not a switch jump instruction.".format(self._ea))

        return results

    def _build_map(self, results):
        switch_map = {}
        for cases, target in zip(results.cases, results.targets):
            for case in cases:
                switch_map[case] = target

        return switch_map

    @property
    def targets(self):
        """Switch Targets"""
        return list(self._map.values())

    @property
    def cases(self):
        """Switch Cases"""
        return list(self._map.keys())

    @property
    def pairs(self):
        """(case, target) pairs"""
        return iter(self._map.items())

    def __iter__(self):
        """Iterate switch cases."""
        return iter(self._map.keys())

    def __getitem__(self, case):
        """switch[case] -> target"""
        return self._map[case]

    def get_cases(self, target):
        """switch.get_cases(target) -> [case]"""
        if target in self.targets:
            return self._reverse_map[target]

        raise KeyError("Target 0x{:08X} does not exist.".format(target))


def is_switch(ea):
    try:
        switch = Switch(ea)
        return True
    except exceptions.SarkNotASwitch:
        return False


import idc
import idaapi
import idautils
from .function import functions
from .line import lines
from .. import exceptions

class Comments(object):
    def __init__(self, segment):
        super(Comments, self).__init__()
        self._segment = segment

    @property
    def regular(self):
        return idaapi.get_segment_cmt(self._segment, False)

    @regular.setter
    def regular(self, comment):
        idaapi.set_segment_cmt(self._segment, comment, False)

    @property
    def repeat(self):
        return idaapi.get_segment_cmt(self._segment, True)

    @repeat.setter
    def repeat(self, comment):
        idaapi.set_segment_cmt(self._segment, comment, True)


class SegmentPermissions(object):
    def __init__(self, segment_t):
        super(SegmentPermissions, self).__init__()
        self._segment = segment_t

    @property
    def x(self):
        return bool(self._segment.perm & idaapi.SEGPERM_EXEC)

    @x.setter
    def x(self, value):
        if value:
            self._segment.perm |= idaapi.SEGPERM_EXEC

        else:
            self._segment.perm &= ~idaapi.SEGPERM_EXEC

    @property
    def w(self):
        return bool(self._segment.perm & idaapi.SEGPERM_WRITE)

    @w.setter
    def w(self, value):
        if value:
            self._segment.perm |= idaapi.SEGPERM_WRITE

        else:
            self._segment.perm &= ~idaapi.SEGPERM_WRITE

    @property
    def r(self):
        return bool(self._segment.perm & idaapi.SEGPERM_READ)

    @r.setter
    def r(self, value):
        if value:
            self._segment.perm |= idaapi.SEGPERM_READ

        else:
            self._segment.perm &= ~idaapi.SEGPERM_READ

    execute = x
    write = w
    read = r

    def __str__(self):
        return "".join(("R" if self.r else "", "W" if self.w else "", "X" if self.x else ""))

    def __repr__(self):
        return "<SegmentPermissions(read={}, write={}, execute={})>".format(self.r, self.w, self.x)


class Segment(object):
    BITNESS_TO_BITS = {
        0: 16,
        1: 32,
        2: 64,
    }

    BITS_TO_BITNESS = {
        16: 0,
        32: 1,
        64: 2,
    }

    class UseCurrentAddress(object):
        pass

    def __init__(self, ea=UseCurrentAddress, name=None, index=None, segment_t=None):
        """Wrapper around IDA segments.

        There are 3 ways to get a segment - by name, ea or index. Only use one.

        Args:
            ea - address in the segment
            name - name of the segment
            index - index of the segment
        """
        if sum((ea not in (self.UseCurrentAddress, None), name is not None, index is not None,
                segment_t is not None,)) > 1:
            raise ValueError((
                                 "Expected only one (ea, name, index or segment_t)."
                                 " Got (ea={!r}, name={!r}, index={!r}, segment_t={!r})"
                             ).format(ea,
                                      name,
                                      index,
                                      segment_t))


        elif segment_t is not None:
            seg = segment_t

        elif name is not None:
            seg = idaapi.get_segm_by_name(name)

        elif index is not None:
            seg = idaapi.getnseg(index)

        elif ea == self.UseCurrentAddress:
            seg = idaapi.getseg(idc.here())

        elif ea is None:
            raise ValueError("`None` is not a valid address. To use the current screen ea, "
                             "use `Function(ea=Function.UseCurrentAddress)` or supply no `ea`.")

        else:
            seg = idaapi.getseg(ea)

        self._segment = seg

    @property
    def segment_t(self):
        return self._segment

    @property
    def permissions(self):
        """Segment permissions.

        Can be used to get or set the segment permissions (r/w/x).

        Returns:
            A `SegmentPermissions` object.
        """
        return SegmentPermissions(self.segment_t)

    @property
    def start_ea(self):
        return self._segment.start_ea

    ea = start_ea

    @property
    def end_ea(self):
        return self._segment.end_ea

    @property
    def type(self):
        return idaapi.segtype(self.ea)

    @property
    def name(self):
        return idaapi.get_segm_name(self.segment_t)

    @name.setter
    def name(self, name):
        idaapi.set_segm_name(self.segment_t, name)

    @property
    def class_(self):
        return idaapi.get_segm_class(self.segment_t)

    @class_.setter
    def class_(self, value):
        idaapi.set_segm_class(self.segment_t, value)

    @property
    def functions(self):
        """Iterate all functions in the segment."""
        return functions(self.start_ea, self.end_ea)

    @property
    def lines(self):
        """Iterate all lines in the segment."""
        return lines(self.start_ea, self.end_ea)

    @property
    def bitness(self):
        """Segment's Bitness.

        Can be 16, 32 or 64.
        """
        return self.BITNESS_TO_BITS[self.segment_t.bitness]

    @bitness.setter
    def bitness(self, bits):
        try:
            self.segment_t.bitness = self.BITS_TO_BITNESS[bits]
        except KeyError:
            raise exceptions.InvalidBitness("Got {}. Expecting 16, 32 or 64.".format(bits))

    @property
    def next(self):
        """Get the next segment."""
        seg = Segment(segment_t=idaapi.get_next_seg(self.ea))

        if seg.ea <= self.ea:
            raise exceptions.NoMoreSegments("This is the last segment. No segments exist after it.")

        return seg

    @property
    def prev(self):
        """Get the previous segment."""
        seg = Segment(segment_t=idaapi.get_prev_seg(self.ea))

        if seg.ea >= self.ea:
            raise exceptions.NoMoreSegments("This is the first segment. no segments exist before it.")

        return seg

    @property
    def comments(self):
        return Comments(self.segment_t)

    @property
    def size(self):
        return self.end_ea - self.start_ea

    def __repr__(self):
        return ("<Segment(ea=0x{:08X},"
                " name={!r},"
                " size=0x{:08X},"
                " permissions={!r},"
                " bitness={})>").format(self.ea,
                                        self.name,
                                        self.size,
                                        str(self.permissions),
                                        self.bitness)


def segments(seg_type=None):
    """Iterate segments based on type

        Args:
            seg_type: type of segment e.g. SEG_CODE

        Returns:
            iterator of `Segment` objects. if seg_type is None , returns all segments
            otherwise returns only the relevant ones
    """

    for index in range(idaapi.get_segm_qty()):
        seg = Segment(index=index)
        if (seg_type is None) or (seg.type == seg_type):
            yield Segment(index=index)



from typing import Iterable

import idaapi
import idautils
import idc
from .base import get_func, demangle
from ..core import set_name, get_ea, fix_addresses, is_same_function, add_func
from .line import Line
from .xref import Xref
from ..ui import updates_ui
from .. import exceptions


class Comments(object):
    """IDA Function Comments

    Provides easy access to all types of comments for an IDA Function.
    """

    def __init__(self, function):
        self._function = function

    def __bool__(self):
        return any((self.regular, self.repeat,))

    @property
    def regular(self):
        """Function Comment"""
        return idaapi.get_func_cmt(self._function._func, False)

    @regular.setter
    def regular(self, comment):
        idaapi.set_func_cmt(self._function._func, comment, False)

    @property
    def repeat(self):
        """Repeatable Function Comment"""
        return idaapi.get_func_cmt(self._function._func, True)

    @repeat.setter
    def repeat(self, comment):
        idaapi.set_func_cmt(self._function._func, comment, True)

    def __repr__(self):
        return ("Comments("
                "func={name},"
                " reqular={regular},"
                " repeat={repeat})").format(
            name=self._function.name,
            regular=repr(self.regular),
            repeat=repr(self.repeat))


class FunctionFlagsMixin(object):
    """ Mixin to add convenience checks for the function flags provided by IDA.

    IDA SDK documentation for the flags is found at:
    https://www.hex-rays.com/products/ida/support/sdkdoc/group___f_u_n_c__.html
    """

    flags = None

    @property
    def is_noret(self):
        """ Function doesn't return """
        return bool(self.flags & idaapi.FUNC_NORET)  # 0x00000001

    @property
    def is_far(self):
        """ Is a far function. """
        return bool(self.flags & idaapi.FUNC_FAR)  # 0x00000002

    @property
    def is_library(self):
        """ Is a library function. """
        return bool(self.flags & idaapi.FUNC_LIB)  # 0x00000004

    @property
    def is_static(self):
        """ Is a static function. """
        return bool(self.flags & idaapi.FUNC_STATICDEF)  # 0x00000008

    @property
    def is_frame(self):
        """ Function uses frame pointer (BP) """
        return bool(self.flags & idaapi.FUNC_FRAME)  # 0x00000010

    @property
    def is_user_far(self):
        """ User has specified far-ness of the function. """
        return bool(self.flags & idaapi.FUNC_USERFAR)  # 0x00000020

    @property
    def is_hidden(self):
        """ A hidden function chunk. """
        return bool(self.flags & idaapi.FUNC_HIDDEN)  # 0x00000040

    @property
    def is_thunk(self):
        """ Thunk (jump) function. """
        return bool(self.flags & idaapi.FUNC_THUNK)  # 0x00000080

    @property
    def is_bottom_bp(self):
        """ BP points to the bottom of the stack frame. """
        return bool(self.flags & idaapi.FUNC_BOTTOMBP)  # 0x00000100

    @property
    def is_noret_pending(self):
        """ Function 'non-return' analysis must be performed. """
        return bool(self.flags & idaapi.FUNC_NORET_PENDING)  # 0x00200

    @property
    def is_sp_ready(self):
        """ SP-analysis has been performed. """
        return bool(self.flags & idaapi.FUNC_SP_READY)  # 0x00000400

    @property
    def is_purged_ok(self):
        """ 'argsize' field has been validated. """
        return bool(self.flags & idaapi.FUNC_PURGED_OK)  # 0x00004000

    @property
    def is_tail(self):
        """ This is a function tail. """
        return bool(self.flags & idaapi.FUNC_TAIL)  # 0x00008000


class Function(FunctionFlagsMixin):
    """IDA Function

    Provides easy access to function related APIs in IDA.
    """

    class UseCurrentAddress(object):
        """
        This is a filler object to replace `None` for the EA.
        In many cases, a programmer can accidentally initialize the
        `Function` object with `ea=None`, resulting in the current address.
        Usually, this is not the desired outcome. This object resolves this issue.
        """
        pass

    def __init__(self, ea=UseCurrentAddress, name=None):
        if name is not None and ea != self.UseCurrentAddress:
            raise ValueError(("Either supply a name or an address (ea). "
                              "Not both. (ea={!r}, name={!r})").format(ea, name))

        elif name is not None:
            ea = idc.get_name_ea_simple(name)
            if ea == idc.BADADDR:
                raise exceptions.SarkNoFunction(
                    "The supplied name does not belong to an existing function. "
                    "(name = {!r})".format(name))

        elif ea == self.UseCurrentAddress:
            ea = idc.here()

        elif ea is None:
            raise ValueError("`None` is not a valid address. To use the current screen ea, "
                             "use `Function(ea=Function.UseCurrentAddress)` or supply no `ea`.")

        elif isinstance(ea, Line):
            ea = ea.ea
        self._func = get_func(ea)
        self._comments = Comments(self)

    @staticmethod
    def is_function(ea=UseCurrentAddress):
        try:
            Function(ea)
            return True
        except exceptions.SarkNoFunction:
            return False

    @staticmethod
    def create(ea=UseCurrentAddress):
        if ea == Function.UseCurrentAddress:
            ea = idc.here()

        if Function.is_function(ea):
            raise exceptions.SarkFunctionExists("Function already exists")

        if not add_func(ea):
            raise exceptions.SarkAddFunctionFailed("Failed to add function")

        return Function(ea)

    @property
    def comments(self):
        """Comments"""
        return self._comments

    def __eq__(self, other):
        try:
            return self.start_ea == other.start_ea
        except AttributeError:
            return False

    def __hash__(self):
        return self.start_ea

    @property
    def lines(self) -> Iterable[Line]:
        """Get all function lines."""
        return iter_function_lines(self._func)

    @property
    def start_ea(self) -> int:
        """Start Address"""
        return self._func.start_ea

    # Alias for `start_ea` for increased guessability and less typing.
    ea = start_ea

    @property
    def end_ea(self):
        """End Address

        Note that taking all the lines between `start_ea` and `end_ea` does not guarantee
        that you get all the lines in the function. To get all the lines, use `.lines`.
        """
        return self._func.end_ea

    @property
    def flags(self):
        """The function flags.

        See `idaapi.FUNC_*` constants.
        """
        return self._func.flags

    @property
    def xrefs_from(self):
        """Xrefs from the function.

        This includes the xrefs from every line in the function, as `Xref` objects.
        Xrefs are filtered to exclude code references that are internal to the function. This
        means that every xrefs to the function's code will NOT be returned (yet, references
        to the function's data will be returnd). To get those extra xrefs, you need to iterate
        the function's lines yourself.
        """
        for line in self.lines:
            for xref in line.xrefs_from:
                if xref.type.is_flow:
                    continue

                if xref.to in self and xref.iscode:
                    continue

                yield xref

    @property
    def calls_from(self):
        return (xref for xref in self.xrefs_from if xref.type.is_call)

    @property
    def drefs_from(self):
        """Destination addresses of data xrefs from this function."""
        for line in self.lines:
            for ea in line.drefs_from:
                yield ea

    @property
    def crefs_from(self):
        """Destination addresses of code xrefs from this function."""
        for line in self.lines:
            for ea in line.crefs_from:
                yield ea

    @property
    def xrefs_to(self):
        """Xrefs to the function.

        This only includes references to that function's start address.
        """
        return map(Xref, idautils.XrefsTo(self.start_ea))

    @property
    def drefs_to(self):
        """Source addresses of data xrefs to this function."""
        return idautils.DataRefsTo(self.start_ea)

    @property
    def crefs_to(self):
        """Source addresses of code xrefs to this function."""
        return idautils.CodeRefsTo(self.start_ea, 1)

    @property
    def name(self):
        """Function's Name"""
        return idaapi.get_ea_name(self.start_ea)

    @property
    def demangled(self):
        """Return the demangled name of the function. If none exists, return `.name`"""
        return demangle(self.name)

    @name.setter
    def name(self, name):
        """Set the function name.

        If the name exists, an exception will be raised. To use IDA's name counting use
        `.set_name(desired_name, anyway=True)`.
        """
        self.set_name(name)

    def set_name(self, name, anyway=False):
        """Set Function Name.

        Default behavior throws an exception when setting to a name that already exists in
        the IDB. to make IDA automatically add a counter to the name (like in the GUI,)
        use `anyway=True`.

        Args:
            name: Desired name.
            anyway: `True` to set anyway.
        """
        set_name(self.start_ea, name, anyway=anyway)

    def __repr__(self):
        return 'Function(name="{}", addr=0x{:08X})'.format(self.name, self.start_ea)

    def __contains__(self, item):
        """Is an item contained (its EA is in) the function."""
        # If the item has an EA, use it. If not, use the item itself assuming it is an EA.
        ea = getattr(item, "ea", item)

        return is_same_function(ea, self.ea)

    @property
    def frame_size(self):
        return idaapi.get_frame_size(self._func)

    @property
    def color(self):
        """Function color in IDA View"""
        color = idc.get_color(self.ea, idc.CIC_FUNC)
        if color == 0xFFFFFFFF:
            return None

        return color

    @color.setter
    @updates_ui
    def color(self, color):
        """Function Color in IDA View.

        Set color to `None` to clear the color.
        """
        if color is None:
            color = 0xFFFFFFFF

        idc.set_color(self.ea, idc.CIC_FUNC, color)

    @property
    def has_name(self):
        return Line(self.start_ea).has_name

    @property
    def func_t(self):
        return self._func

    @property
    def signature(self):
        '''The C signature of the function.'''
        return idc.get_type(self.start_ea)

    @signature.setter
    def signature(self, c_signature):
        success = idc.SetType(self.start_ea, c_signature)
        if not success:
            raise exceptions.SetTypeFailed(self.start_ea, c_signature)

    @property
    def tinfo(self):
        '''The tinfo of the function type'''
        return idc.get_tinfo(self.start_ea)

    @tinfo.setter
    def tinfo(self, tinfo):
        success = idc.apply_type(self.start_ea, tinfo)
        if not success:
            raise exceptions.SetTypeFailed(self.start_ea, tinfo)


def iter_function_lines(func_ea) -> Iterable[Line]:
    """Iterate the lines of a function.

    Args:
        func_ea (idaapi.func_t, int): The function to iterate.

    Returns:
        Iterator over all the lines of the function.
    """
    for line in idautils.FuncItems(get_ea(func_ea)):
        yield Line(line)


def functions(start=None, end=None):
    """Get all functions in range.

    Args:
        start: Start address of the range. Defaults to IDB start.
        end: End address of the range. Defaults to IDB end.

    Returns:
        This is a generator that iterates over all the functions in the IDB.
    """
    start, end = fix_addresses(start, end)

    for func_t in idautils.Functions(start, end):
        yield Function(func_t)



import idaapi
import idc
import ida_search
import string
from . import exceptions


def get_func(func_ea):
    """get_func(func_t or ea) -> func_t

    Take an IDA function (``idaapi.func_t``) or an address (EA) and return
    an IDA function object.

    Use this when APIs can take either a function or an address.

    Args:
        func_ea: ``idaapi.func_t`` or ea of the function.

    Returns:
        An ``idaapi.func_t`` object for the given address. If a ``func_t`` is
        provided, it is returned.
    """
    if isinstance(func_ea, idaapi.func_t):
        return func_ea
    func = idaapi.get_func(func_ea)
    if func is None:
        raise exceptions.SarkNoFunction("No function at 0x{:08X}".format(func_ea))

    return func


def get_ea(func_ea):
    """get_ea(func_t or ea) -> ea

    Same as `get_func`, but returns the EA.

    Args:
        func_ea: `idaapi.func_t` or EA.

    Returns:
        The ea.
    """
    if isinstance(func_ea, idaapi.func_t):
        return func_ea.start_ea
    return func_ea


def add_func(func_ea,func_end=idaapi.BADADDR):
    return idaapi.add_func(func_ea,func_end)

def is_string_printable(string_):
    """Check if a string is printable"""
    return set(string_) - set(string.printable)


def string_to_query(string_):
    if is_string_printable(string_):
        return '"{}"'.format(string_)

    return " ".join(char.encode("hex") for char in string_)


def iter_find_string(query, start=None, end=None, down=True):
    query = string_to_query(query)
    return iter_find_query(query, start, end, down)


def iter_find_query(query, start=None, end=None, down=True):
    start, end = fix_addresses(start, end)

    if down:
        direction = ida_search.SEARCH_DOWN
    else:
        direction = ida_search.SEARCH_UP

    current = ida_search.find_binary(start, end, query, 16, direction)
    while current < end:
        yield current
        current = ida_search.find_binary(current + 1, end, query, 16, direction)


def fix_addresses(start=None, end=None):
    """Set missing addresses to start and end of IDB.

    Take a start and end addresses. If an address is None or `BADADDR`,
    return start or end addresses of the IDB instead.

    Args
        start: Start EA. Use `None` to get IDB start.
        end:  End EA. Use `None` to get IDB end.

    Returns:
        (start, end)
    """
    if start in (None, idaapi.BADADDR):
        start = idaapi.cvar.inf.min_ea

    if end in (None, idaapi.BADADDR):
        end = idaapi.cvar.inf.max_ea

    return start, end


def set_name(address, name, anyway=False):
    """Set the name of an address.

    Sets the name of an address in IDA.
    If the name already exists, check the `anyway` parameter:

        True - Add `_COUNTER` to the name (default IDA behaviour)
        False - Raise an `exceptions.SarkErrorNameAlreadyExists` exception.


    Args
        address: The address to rename.
        name: The desired name.
        anyway: Set anyway or not. Defualt ``False``.
    """
    success = idaapi.set_name(address, name, idaapi.SN_NOWARN | idaapi.SN_NOCHECK)
    if success:
        return

    if anyway:
        success = idaapi.force_name(address, name)
        if success:
            return

        raise exceptions.SarkSetNameFailed("Failed renaming 0x{:08X} to {!r}.".format(address, name))

    raise exceptions.SarkErrorNameAlreadyExists(
        "Can't rename 0x{:08X}. Name {!r} already exists.".format(address, name))


def is_same_function(ea1, ea2):
    """Are both addresses in the same function?"""
    func1 = idaapi.get_func(ea1)
    func2 = idaapi.get_func(ea2)
    # This is bloated code. `None in (func1, func2)` will not work because of a
    # bug in IDAPython in the way functions are compared.
    if any(func is None for func in (func1, func2)):
        return False

    return func1.start_ea == func2.start_ea


def get_name_or_address(ea):
    name = idaapi.get_ea_name(ea, idaapi.GN_VISIBLE)
    if name:
        name = repr(name)
    else:
        name = "0x{:08X}".format(ea)

    return name


def get_native_size():
    """Get the native word size in normal 8-bit bytes."""
    info = idaapi.get_inf_structure()
    if info.is_64bit():
        return 8
    elif info.is_32bit():
        return 4
    else:
        return 2


def get_fileregion_offset(ea):
    file_offset = idaapi.get_fileregion_offset(ea)
    if file_offset == -1:
        raise exceptions.NoFileOffset("Address 0x{:08X} is not mapped to any file offset.".format(ea))

    return file_offset

def is_function(ea):
    try:
        get_func(ea)
        return True
    except exceptions.SarkNoFunction:
        return False


def is_signed(number, size=None):
    if not size:
        size = get_native_size()
    return number & (1 << ((8 * size) - 1))





def is_in_ida():
    try:
        import idaapi

        return True
    except ImportError:
        return False


# Since some of our code can be used outside of IDA, namely the `plumbing` module
# when used in the codecs proxy, we want to allow importing specific modules outside
# IDA.
if is_in_ida():
    from . import (core,
                   code,
                   exceptions,
                   structure,
                   codeblock,
                   data,
                   debug,
                   enum,
                   ui,
                   graph)

    import idaapi
    idaapi.require('sark.code')
    idaapi.require('sark.core')
    idaapi.require('sark.exceptions')
    idaapi.require('sark.graph')
    idaapi.require('sark.structure')
    idaapi.require('sark.codeblock')
    idaapi.require('sark.data')
    idaapi.require('sark.debug')
    idaapi.require('sark.enum')
    idaapi.require('sark.ui')

    from .code import *
    from .codeblock import CodeBlock, get_nx_graph, get_block_start, FlowChart, codeblocks
    from .data import read_ascii_string, get_string
    from .core import set_name, is_function
    from .enum import Enum, enums, add_enum, remove_enum



from collections import namedtuple
import idaapi


class Registers(object):
    """
    Wrapper around IDA's debug registers.

    Enables easy querying of the debug registers API (`idaapi.dbg_get_registers`) to get
    register info, such as the names of the instruction pointer and the stack pointer.

    Usage (on x86):
        >>> print(Registers().ip.name)
        eip

        >>> print(Registers().sp.name)
        esp
    """
    REGISTER_READONLY = 0x001  # the user can't modify the current value of this register
    REGISTER_IP = 0x002  # instruction pointer
    REGISTER_SP = 0x004  # stack pointer
    REGISTER_FP = 0x008  # frame pointer
    REGISTER_ADDRESS = 0x010  # may contain an address
    REGISTER_CS = 0x020  # code segment
    REGSITER_SS = 0x040  # stack segment
    REGISTER_NOLF = 0x080  # displays this register without returning to the next line
    REGISTER_CUSTFMT = 0x100  # allowing the next register to be displayed to its right (on the same line)
    # register should be displayed using a custom data format.

    RegisterInfo = namedtuple("RegisterInfo", "name, flags, cls, dtyp, bit_strings, bit_strings_default_mask")

    def __init__(self):
        reg_infos = idaapi.dbg_get_registers()
        if not reg_infos:
            raise RuntimeError("Debugger not present.")

        self._reg_infos = [self.RegisterInfo(*reg_info) for reg_info in reg_infos]

    def get_by_flags(self, flags):
        """Iterate all register infos matching the given flags."""
        for reg in self._reg_infos:
            if reg.flags & flags == flags:
                yield reg

    def get_single_by_flags(self, flags):
        """Get the register info matching the flag. Raises ValueError if more than one are found."""
        regs = list(self.get_by_flags(flags))
        if len(regs) != 1:
            raise ValueError("Flags do not return unique resigter. {!r}", regs)

        return regs[0]

    @property
    def ip(self):
        """Instruction Pointer"""
        return self.get_single_by_flags(self.REGISTER_IP)

    pc = ip

    @property
    def sp(self):
        """Stack Pointer"""
        return self.get_single_by_flags(self.REGISTER_SP)

    @property
    def fp(self):
        """Frame Pointer"""
        return self.get_single_by_flags(self.REGISTER_FP)


import idaapi
import idc
from threading import RLock
import itertools
import wrapt
import traceback

if idaapi.is_idaq():
    # Only load Qt if we're in IDAQ.
    # This should allow running in IDAT.
    from .qt import MenuManager
    from PyQt5 import QtGui, QtWidgets

def ask_color(initial=None):
    if initial is not None:
        color = QtGui.QColor(initial & 0xFF, (initial >> 8) & 0xFF, (initial >> 16) & 0xFF)
        qcolor_dialog = QtWidgets.QColorDialog(color)

    else:
        qcolor_dialog = QtWidgets.QColorDialog()

    qcolor = qcolor_dialog.getColor()

    if not qcolor.isValid:
        return None

    return (qcolor.blue() << 16) | (qcolor.green() << 8) | (qcolor.red() << 0)


class Update(object):
    """
    A context manager that refreshes the UI on `__exit__`.

    When nested, only the topmost context manager can update the UI.
    This is used to make sure that UI-heavy code does not update the
    UI until it is finished.

    Note that this does not prevent updates via other APIs, so be sure
    to use this and the `updates_ui` decorator.
    """
    LOCK = RLock()

    def __enter__(self):
        self.LOCK.acquire(blocking=False)

    def __exit__(self, exc_type, exc_val, exc_tb):
        can_update = self.LOCK._is_owned()

        if can_update:
            idaapi.refresh_idaview_anyway()

        self.LOCK.release()


@wrapt.decorator
def updates_ui(wrapped, instance, args, kwargs):
    """Refresh UI on return."""
    with Update():
        return wrapped(*args, **kwargs)


class BasicNodeHandler(object):
    """Basic Node Handler

    This is the base class for all node handlers (for NXGraph).
    It implements usable defaults for all required events.

    When subclassing, simply replace the events you want to modify.
    """

    def on_get_text(self, value, attrs):
        """Get the text to display on the node.

        Args:
            value: The value of the current node.
            attrs (dict): The node's attributes.

        Returns:
            str: The text to display.
        """
        return str(value)

    def on_click(self, value, attrs):
        """Action to perform on click.

        Args:
            value: The value of the current node.
            attrs (dict): The node's attributes.

        Returns:
            ``True`` to accept the click, ``False`` to ignore it.
        """
        return False

    def on_double_click(self, value, attrs):
        """Action to perform on double click.

        Args:
            value: The value of the current node.
            attrs (dict): The node's attributes.

        Returns:
            ``True`` to accept the click, ``False`` to ignore it.
        """
        return False

    def on_hint(self, value, attrs):
        """Hint to show.

        Args:
            value: The value of the current node.
            attrs (dict): The node's attributes.

        Returns:
            The hint to show.
        """
        return None

    def on_bg_color(self, value, attrs):
        """Background color.

        Args:
            value: The value of the current node.
            attrs (dict): The node's attributes.

        Returns:
            ``None`` for default color, otherwise the color as a number.
        """
        return attrs.get(NXGraph.BG_COLOR, None)

    def on_frame_color(self, value, attrs):
        """Frame color.

        Args:
            value: The value of the current node.
            attrs (dict): The node's attributes.

        Returns:
            ``None`` for default color, otherwise the color as a number.
        """
        return attrs.get(NXGraph.FRAME_COLOR, None)


class AddressNodeHandler(BasicNodeHandler):
    """Address Node Handler

    Used to display addresses.
    In addition to the default functionality:
        1. Shows the name of an address (hex value if no name exists) instead of
            just the number;
        2. On double-click, jumps to the address clicked.
    """

    def on_get_text(self, value, attrs):
        name = idaapi.get_ea_name(value)
        demangle = getattr(idaapi, 'demangle_name2', idaapi.demangle_name)
        name = demangle(name, 0) or name
        return name or "0x{:08X}".format(value)

    def on_double_click(self, value, attrs):
        idaapi.jumpto(value)
        return False


class NXGraph(idaapi.GraphViewer):
    """NetworkX Graph Viewer

    A utility class for displaying NetworkX graphs inside IDA with ease.

    When showing a graph, the nodes and edges are iterated to create the graph structure.
    For every node, a "handler" is used to get display parameters:
        - Text to display
        - Background and Frame colors
        - Hint
        - Actions on click and double-click

    Handlers can be specified in one of 2 ways:
        1. By specifying the `handler` parameter to the constructor;
        2. By setting the `NXGraph.HANDLER` attribute of a specific node:

            >>> my_graph.nodes[my_node][NXGraph.HANDLER] = MyCustomHandler()

    Two other useful attribute are `NXGraph.BG_COLOR` and `NXGraph.FRAME_COLOR` that allow
    specifying colors for your nodes. If not provided, the default color will be used.
    Note that the handler is responsible for using those attributes, and can therefore
    modify their behaviour or ignore them completely.

    for more information about handlers see the `BasicNodeHandler` class.

    To display a graph, use:

        >>> viewer = sark.ui.NXGraph(my_graph, title="My Graph")
        >>> viewer.Show()


    To make the graph easier on the eye, node text is padded, adding empty space around it.
    `PADDING` is the default amount of padding to use in all directions.
    `PADDING_WIDTH` scales the padding width.
    To change the padding, simply provide the constructor with a padding to use.
    """
    PAD_WIDTH = 3
    PADDING = 1
    HANDLER = "HANDLER"
    BG_COLOR = "BG_COLOR"
    FRAME_COLOR = "FRAME_COLOR"
    DEFAULT_HANDLER = BasicNodeHandler()

    def __init__(self, graph, title="GraphViewer", handler=None, padding=PADDING):
        """Initialize the graph viewer.

        To avoid bizarre IDA errors (crashing when creating 2 graphs with the same title,)
        a counter is appended to the title (similar to "Hex View-1".)

        Args:
            graph: A NetworkX graph to display.
            title: The graph title.
            handler: The default node handler to use when accessing node data.
        """
        title = self._make_unique_title(title)

        idaapi.GraphViewer.__init__(self, title)

        self._graph = graph

        if handler is None:
            handler = self.DEFAULT_HANDLER

        # Here we make sure the handler is an instance of `BasicNodeHandler` or inherited
        # types. While generally being bad Python practice, we still need it here as an
        # invalid handler can cause IDA to crash.
        if not isinstance(handler, BasicNodeHandler):
            raise TypeError("Node handler must inherit from `BasicNodeHandler`.")

        self._default_handler = handler
        self._padding = padding

    def _pad(self, text):
        """Pad the text."""
        top_bottom = ("\n" * self._padding) + " "
        right_left = " " * self._padding * self.PAD_WIDTH
        return top_bottom + right_left + text + right_left + top_bottom

    def _make_unique_title(self, title):
        """Make the title unique.

        Adds a counter to the title to prevent duplicates.

        Prior to IDA 6.8, two graphs with the same title could crash IDA.
        This has been fixed (https://www.hex-rays.com/products/ida/6.8/index.shtml).
        The code will not change for support of older versions and as it is
        more usable this way.
        """
        unique_title = title

        for counter in itertools.count():
            unique_title = "{}-{}".format(title, counter)
            if not idaapi.find_widget(unique_title):
                break

        return unique_title

    def _get_handler(self, node_id):
        """Get the handler of a given node."""
        handler = self._get_attrs(node_id).get(self.HANDLER, self._default_handler)

        # Here we make sure the handler is an instance of `BasicNodeHandler` or inherited
        # types. While generally being bad Python practice, we still need it here as an
        # invalid handler can cause IDA to crash.
        if not isinstance(handler, BasicNodeHandler):
            idaapi.msg(("Invalid handler for node {}: {}. All handlers must inherit from"
                        "`BasicNodeHandler`.").format(node_id, handler))
            handler = self._default_handler
        return handler

    def _get_attrs(self, node_id):
        """Get the node's attributes"""
        return self._graph.nodes[self[node_id]]

    def _get_handling_triplet(self, node_id):
        """_get_handling_triplet(node_id) -> (handler, value, attrs)"""
        handler = self._get_handler(node_id)
        value = self[node_id]
        attrs = self._get_attrs(node_id)

        return handler, value, attrs

    def _OnNodeInfo(self, node_id):
        """Sets the node info based on its attributes."""
        handler, value, attrs = self._get_handling_triplet(node_id)
        frame_color = handler.on_frame_color(value, attrs)

        node_info = idaapi.node_info_t()

        if frame_color is not None:
            node_info.frame_color = frame_color

        flags = node_info.get_flags_for_valid()

        self.SetNodeInfo(node_id, node_info, flags)

    def update_node_info(self):
        """Sets the node info for all nodes."""
        for node_id, node in enumerate(self):
            self._OnNodeInfo(node_id)

    def OnGetText(self, node_id):
        handler, value, attrs = self._get_handling_triplet(node_id)
        self._OnNodeInfo(node_id)
        return (self._pad(handler.on_get_text(value, attrs)), handler.on_bg_color(value, attrs))

    def Show(self):
        if not idaapi.GraphViewer.Show(self):
            return False

        return True

    def OnRefresh(self):
        self.Clear()

        # Compatibility between NetworkX 1.x and 2.x
        try:
            graph_nodes_iter = self._graph.nodes()
            graph_edges_iter = self._graph.edges()
        except AttributeError:
            graph_nodes_iter = self._graph.nodes_iter()
            graph_edges_iter = self._graph.edges_iter()

        node_ids = {node: self.AddNode(node) for node in graph_nodes_iter}

        for frm, to in graph_edges_iter:
            self.AddEdge(node_ids[frm], node_ids[to])

        self.update_node_info()

        return True

    def OnActivate(self):
        # Refresh on every activation to keep the graph up to date.
        self.Refresh()
        return True

    def OnDeactivate(self):
        pass

    def OnDblClick(self, node_id):
        handler, value, attrs = self._get_handling_triplet(node_id)
        return handler.on_double_click(value, attrs)

    def OnClick(self, node_id):
        handler, value, attrs = self._get_handling_triplet(node_id)
        return handler.on_click(value, attrs)

    def OnHint(self, node_id):
        handler, value, attrs = self._get_handling_triplet(node_id)
        return handler.on_hint(value, attrs)


class ActionHandler(idaapi.action_handler_t):
    """A wrapper around `idaapi.action_handler_t`.

    The class simplifies the creation of UI actions in IDA >= 6.7.

    To create an action, simply create subclass and override the relevant fields
    and register it::

        class MyAction(ActionHandler):
            TEXT = "My Action"
            HOTKEY = "Alt+Z"

            def _activate(self, ctx):
                idaapi.msg("Activated!")

        MyAction.register()

    Additional Documentation:
        Introduction to `idaapi.action_handler_t`:
            http://www.hexblog.com/?p=886

        Return values for update (from the SDK):
            AST_ENABLE_ALWAYS     // enable action and do not call action_handler_t::update() anymore
            AST_ENABLE_FOR_IDB    // enable action for the current idb. Call action_handler_t::update() when a database is opened/closed
            AST_ENABLE_FOR_WIDGET // enable action for the current widget. Call action_handler_t::update() when a form gets/loses focus
            AST_ENABLE            // enable action - call action_handler_t::update() when anything changes

            AST_DISABLE_ALWAYS    // disable action and do not call action_handler_t::action() anymore
            AST_DISABLE_FOR_IDB   // analog of ::AST_ENABLE_FOR_IDB
            AST_DISABLE_FOR_WIDGET// analog of ::AST_ENABLE_FOR_WIDGET
            AST_DISABLE           // analog of ::AST_ENABLE
    """
    NAME = None
    TEXT = "Default. Replace me!"
    HOTKEY = ""
    TOOLTIP = ""
    ICON = -1

    @classmethod
    def get_name(cls):
        """Return the name of the action.

        If a name has not been set (using the `Name` class variable), the
        function generates a name based on the class name and id.
        :return: action name
        :rtype: str
        """
        if cls.NAME is not None:
            return cls.NAME

        return "{}:{}".format(cls.__name__, id(cls))

    @classmethod
    def get_desc(cls):
        """Get a descriptor for this handler."""
        name = cls.get_name()
        text = cls.TEXT
        handler = cls()
        hotkey = cls.HOTKEY
        tooltip = cls.TOOLTIP
        icon = cls.ICON
        action_desc = idaapi.action_desc_t(
            name,
            text,
            handler,
            hotkey,
            tooltip,
            icon,
        )
        return action_desc

    @classmethod
    def register(cls):
        """Register the action.

        Each action MUST be registered before it can be used. To remove the action
        use the `unregister` method.
        """
        action_desc = cls.get_desc()

        return idaapi.register_action(action_desc)

    @classmethod
    def unregister(cls):
        """Unregister the action.

        After unregistering the class cannot be used.
        """
        idaapi.unregister_action(cls.get_name())

    def __init__(self):
        idaapi.action_handler_t.__init__(self)

    def activate(self, ctx):
        try:
            self._activate(ctx)
            return 1
        except:
            trace = traceback.format_exc()
            idaapi.msg("Action {!r} failed to activate. Traceback:\n{}".format(self.get_name(), trace))
            return 0

    def update(self, ctx):
        """Update the action.

        Optionally override this function.
        See IDA-SDK for more information.
        """
        return idaapi.AST_ENABLE_ALWAYS

    def _activate(self, ctx):
        """Activate the action.

        This function contains the action code itself. You MUST implement
        it in your class for the action to work.

        Args:
            ctx: The action context passed from IDA.
        """
        raise NotImplementedError()



import idaapi
import idc
from ..core import get_name_or_address


class XrefType(object):
    """Xref Type Wrapper.

    Provides easy to use parsing of xref types.
    All the properties are flag checks on the type value.
    """
    TYPES = {
        0x00: 'Data_Unknown',
        0x01: 'Data_Offset',
        0x02: 'Data_Write',
        0x03: 'Data_Read',
        0x04: 'Data_Text',
        0x05: 'Data_Informational',
        0x10: 'Code_Far_Call',
        0x11: 'Code_Near_Call',
        0x12: 'Code_Far_Jump',
        0x13: 'Code_Near_Jump',
        0x14: 'Code_User',
        0x15: 'Ordinary_Flow'
    }

    def __init__(self, type_):
        self._type = type_

    @property
    def type(self):
        """Xref type, flags excluded."""
        return self._type & idaapi.XREF_MASK

    @property
    def flags(self):
        """Xref flags, type excluded."""
        return self._type ^ self.type

    @property
    def name(self):
        """Name of the xref type."""
        return self.TYPES[self._type]

    def __repr__(self):
        return self.name

    @property
    def is_code(self):
        return self._type & 0x10

    @property
    def is_data(self):
        return not self.is_code

    @property
    def is_unknown(self):
        return self.type == idaapi.fl_U

    @property
    def is_offset(self):
        return self.type == idaapi.dr_O

    @property
    def is_write(self):
        return self.type == idaapi.dr_W

    @property
    def is_read(self):
        return self.type == idaapi.dr_R

    @property
    def is_text(self):
        return self.type == idaapi.dr_T

    @property
    def is_info(self):
        return self.type == idaapi.dr_I

    @property
    def is_far_call(self):
        return self.type == idaapi.fl_CF

    @property
    def is_near_call(self):
        return self.type == idaapi.fl_CN

    @property
    def is_far_jump(self):
        return self.type == idaapi.fl_JF

    @property
    def is_near_jump(self):
        return self.type == idaapi.fl_JN

    @property
    def is_unknown(self):
        return self.type == idaapi.fl_U

    @property
    def is_flow(self):
        return self.type == idaapi.fl_F

    @property
    def is_user(self):
        return self.flags & idaapi.XREF_USER

    @property
    def is_tail(self):
        return self.flags & idaapi.XREF_TAIL

    @property
    def is_base(self):
        return self.flags & idaapi.XREF_BASE

    @property
    def is_call(self):
        return self.is_far_call or self.is_near_call

    @property
    def is_jump(self):
        return self.is_far_jump or self.is_near_jump


class Xref(object):
    """Xref Object

    Provides easy access to xref attributes.
    Most interesting data (xref type) is accessible via the `.type`
    attribute.
    """

    def __init__(self, xref):
        for attr in ['frm', 'to', 'iscode', 'user']:
            setattr(self, attr, getattr(xref, attr))

        self._type = XrefType(xref.type)

    @property
    def type(self):
        return self._type

    def __repr__(self):
        return "<Xref(frm={frm}, to={to}, iscode={iscode}, user={user}, type={type})>".format(
            frm=get_name_or_address(self.frm),
            to=get_name_or_address(self.to),
            iscode=self.iscode,
            user=self.user,
            type=repr(self.type),
        )


import importlib
from . import base

importlib.reload(base)

from . import xref

importlib.reload(xref)

from . import line

importlib.reload(line)

from . import function

importlib.reload(function)

from . import switch

importlib.reload(switch)

from . import instruction

importlib.reload(instruction)

from . import segment

importlib.reload(segment)

from .base import *
from .line import Line, lines
from .function import Function, functions
from .switch import Switch, is_switch
from .segment import Segment, segments


from IPython import display

from .qt import capture_widget, get_window, get_widget

def snap(title=None):
    if title:
        w = get_widget(title)
    else:
        w = get_window()

    display.display_png(capture_widget(w), raw=True)


import idaapi
from . import exceptions
from contextlib import suppress

DEFMASK = idaapi.BADADDR

ENUM_ERROR_MAP = {
    idaapi.ENUM_MEMBER_ERROR_NAME:
        (exceptions.SarkErrorEnumMemberName, "already have member with this name (bad name)"),
    idaapi.ENUM_MEMBER_ERROR_VALUE:
        (exceptions.SarkErrorEnumMemberValue, "already have 256 members with this value"),
    idaapi.ENUM_MEMBER_ERROR_ENUM:
        (exceptions.SarkErrorEnumMemberEnum, "bad enum id"),
    idaapi.ENUM_MEMBER_ERROR_MASK:
        (exceptions.SarkErrorEnumMemberMask, "bad bmask"),
    idaapi.ENUM_MEMBER_ERROR_ILLV:
        (exceptions.SarkErrorEnumMemberIllv, "bad bmask and value combination (~bmask & value != 0)"),
}


def _enum_member_error(err, eid, name, value, bitmask):
    """Format enum member error."""
    exception, msg = ENUM_ERROR_MAP[err]
    enum_name = idaapi.get_enum_name(eid)
    return exception(('add_enum_member(enum="{}", member="{}", value={}, bitmask=0x{:08X}) '
                      'failed: {}').format(
        enum_name,
        name,
        value,
        bitmask,
        msg
    ))


def _get_enum(name):
    """Get an existing enum ID"""
    eid = idaapi.get_enum(name)
    if eid == idaapi.BADADDR:
        raise exceptions.EnumNotFound('Enum "{}" does not exist.'.format(name))
    return eid


def add_enum(name=None, index=None, flags=idaapi.hex_flag(), bitfield=False):
    """Create a new enum.

    Args:
        name: Name of the enum to create.
        index: The index of the enum. Leave at default to append the enum as the last enum.
        flags: Enum type flags.
        bitfield: Is the enum a bitfield.

    Returns:
        An `Enum` object.
    """
    if name is not None:
        with suppress(exceptions.EnumNotFound):
            _get_enum(name)
            raise exceptions.EnumAlreadyExists()

    if index is None or index < 0:
        index = idaapi.get_enum_qty()

    eid = idaapi.add_enum(index, name, flags)

    if eid == idaapi.BADADDR:
        raise exceptions.EnumCreationFailed('Failed creating enum "{}"'.format(name))

    if bitfield:
        idaapi.set_enum_bf(eid, bitfield)

    return Enum(eid=eid)

def remove_enum(name):
    """Delete an enum by name."""
    eid = _get_enum(name)
    idaapi.del_enum(eid)


def _add_enum_member(enum, name, value, bitmask=DEFMASK):
    """Add an enum member."""
    error = idaapi.add_enum_member(enum, name, value, bitmask)

    if error:
        raise _enum_member_error(error, enum, name, value, bitmask)


class EnumComments(object):
    """Enum comments retrieval and manipulation."""

    def __init__(self, eid):
        super(EnumComments, self).__init__()

        self._eid = eid

    @property
    def regular(self):
        return idaapi.get_enum_cmt(self._eid, False)

    @regular.setter
    def regular(self, comment):
        success = idaapi.set_enum_cmt(self._eid, comment, False)
        if not success:
            raise exceptions.CantSetEnumComment("Cant set enum comment.")


    @property
    def repeat(self):
        return idaapi.get_enum_cmt(self._eid, True)

    @repeat.setter
    def repeat(self, comment):
        success = idaapi.set_enum_cmt(self._eid, comment, True)
        if not success:
            raise exceptions.CantSetEnumComment("Cant set enum comment.")

    def __repr__(self):
        return ("EnumComments("
                "name={name!r},"
                " reqular={regular!r},"
                " repeat={repeat!r})").format(
            name=Enum(eid=self._eid).name,
            regular=self.regular,
            repeat=self.repeat, )


class EnumMembers(object):
    """Enum members retrieval and manipulation."""

    def __init__(self, eid):
        super(EnumMembers, self).__init__()
        self._eid = eid

    def __len__(self):
        """Number of members in the enum"""
        return idaapi.get_enum_size(self._eid)

    def __iter__(self):
        """Iterate all members of the enum"""
        return (EnumMember(cid) for cid in _iter_enum_constant_ids(self._eid))

    def add(self, name, value, bitmask=DEFMASK):
        """Add an enum member

        Args:
            name: Name of the member
            value: value of the member
            bitmask: bitmask. Only use if enum is a bitfield.
        """
        _add_enum_member(self._eid, name, value, bitmask)

    def __getitem__(self, name):
        """Get an enum member by name."""
        for enum_member in self:
            if enum_member.name == name:
                return enum_member
        raise KeyError("No member named {!r}".format(name))

    def remove(self, name):
        """Remove an enum member by name"""
        member = self[name]
        serial = member.serial
        value = member.value
        bmask = member.bmask

        success = idaapi.del_enum_member(self._eid, value, serial, bmask)
        if not success:
            raise exceptions.CantDeleteEnumMember("Can't delete enum member {!r}.".format(name))

    def __repr__(self):
        return "<EnumMembers(enum={!r}, members={{{}}})>".format(
            Enum(eid=self._eid).name,
            ", ".join("{member.name!r}: {member.value!r}".format(member=member) for member in self)
        )


class Enum(object):
    """An enum in the IDB"""

    def __init__(self, name=None, eid=None):
        """
        Get an existing enum.

        Only provide one of `name` and `eid`.

        Args:
            name: Name of the enum
            eid: Enum ID
        """
        if None not in (name, eid):
            raise TypeError("Provide only a `name` or an `eid`.")

        self._eid = eid or _get_enum(name)
        self._comments = EnumComments(self._eid)

    @property
    def name(self):
        """Name of the enum"""
        return idaapi.get_enum_name(self.eid)

    @name.setter
    def name(self, name):
        """Set the enum name."""
        success = idaapi.set_enum_name(self.eid, name)
        if not success:
            raise exceptions.CantRenameEnum("Cant rename enum {!r} to {!r}.".format(self.name, name))

    @property
    def width(self):
        """Width of the enum"""
        return idaapi.get_enum_width(self.eid)

    @property
    def comments(self):
        """Enum comments"""
        return self._comments

    @property
    def eid(self):
        """Enum ID"""
        return self._eid

    @property
    def flag(self):
        """Enum flags (bitness, and display type)"""
        return idaapi.get_enum_flag(self.eid)

    @property
    def bitfield(self):
        """Is the enum a bitfield"""
        return idaapi.is_bf(self.eid)

    @bitfield.setter
    def bitfield(self, value):
        success = idaapi.set_enum_bf(self.eid, value)
        if not success:
            raise exceptions.CantSetEnumBitfield()

    @property
    def members(self):
        """Get the enum members."""
        return EnumMembers(self.eid)

    @property
    def is_from_til(self):
        """Is from type library?"""
        return idaapi.is_enum_fromtil(self.eid)

    def __repr__(self):
        return "<Enum(name={!r})>".format(self.name)


class EnumMemberComments(object):
    """Enum member comments retrieval and manipulation."""

    def __init__(self, cid):
        super(EnumMemberComments, self).__init__()

        self._cid = cid

    @property
    def regular(self):
        return idaapi.get_enum_member_cmt(self._cid, False)

    @regular.setter
    def regular(self, comment):
        success = idaapi.set_enum_member_cmt(self._cid, comment, False)
        if not success:
            raise exceptions.CantSetEnumMemberComment("Cant set enum member comment.")

    @property
    def repeat(self):
        return idaapi.get_enum_member_cmt(self._cid, True)

    @repeat.setter
    def repeat(self, comment):
        success = idaapi.set_enum_member_cmt(self._cid, comment, True)
        if not success:
            raise exceptions.CantSetEnumMemberComment("Cant set enum member comment.")

    def __repr__(self):
        enum_member = EnumMember(self._cid)
        return ("EnumMemberComments("
                "name={name!r},"
                " reqular={regular!r},"
                " repeat={repeat!r})").format(
            name="{}.{}".format(enum_member.parent.name, enum_member.name),
            regular=self.regular,
            repeat=self.repeat, )


class EnumMember(object):
    """A member of an enum."""

    def __init__(self, cid):
        super(EnumMember, self).__init__()
        self._cid = cid
        self._comments = EnumMemberComments(self._cid)

    @property
    def cid(self):
        """Get the constant ID"""
        return self._cid

    @property
    def name(self):
        """Get the member name."""
        return idaapi.get_enum_member_name(self.cid)

    @name.setter
    def name(self, name):
        """Set the member name.

        Note that a member name cannot appear in other enums, or generally
        anywhere else in the IDB.
        """
        success = idaapi.set_enum_member_name(self.cid, name)
        if not success:
            raise exceptions.CantRenameEnumMember(
                "Failed renaming {!r} to {!r}. Does the name exist somewhere else?".format(self.name, name))

    @property
    def bmask(self):
        """Get the bitmask"""
        return idaapi.get_enum_member_bmask(self.cid)

    bitmask = bmask

    @property
    def value(self):
        """Get the member value"""
        return idaapi.get_enum_member_value(self.cid)

    @property
    def comments(self):
        """Get the member comments"""
        return self._comments

    @property
    def serial(self):
        """Get the member serial (among members of the same value)."""
        return idaapi.get_enum_member_serial(self.cid)

    @property
    def parent(self):
        """Get the enum holding the member."""
        return Enum(eid=idaapi.get_enum_member_enum(self.cid))

    def __repr__(self):
        return "<EnumMember(name='{}.{}')>".format(self.parent.name, self.name)


def _iter_bitmasks(eid):
    """Iterate all bitmasks in a given enum.

    Note that while `DEFMASK` indicates no-more-bitmasks, it is also a
    valid bitmask value. The only way to tell if it exists is when iterating
    the serials.
    """
    bitmask = idaapi.get_first_bmask(eid)

    yield bitmask

    while bitmask != DEFMASK:
        bitmask = idaapi.get_next_bmask(eid, bitmask)
        yield bitmask


def _iter_enum_member_values(eid, bitmask):
    """Iterate member values with given bitmask inside the enum

    Note that `DEFMASK` can either indicate end-of-values or a valid value.
    Iterate serials to tell apart.
    """
    value = idaapi.get_first_enum_member(eid, bitmask)

    yield value
    while value != DEFMASK:
        value = idaapi.get_next_enum_member(eid, value, bitmask)
        yield value


def _iter_serial_enum_member(eid, value, bitmask):
    """Iterate serial and CID of enum members with given value and bitmask.

    Here only valid values are returned, as `idaapi.BADNODE` always indicates
    an invalid member.
    """
    cid, serial = idaapi.get_first_serial_enum_member(eid, value, bitmask)
    while cid != idaapi.BADNODE:
        yield cid, serial
        cid, serial = idaapi.get_next_serial_enum_member(serial, cid)


def _iter_enum_constant_ids(eid):
    """Iterate the constant IDs of all members in the given enum"""
    for bitmask in _iter_bitmasks(eid):
        for value in _iter_enum_member_values(eid, bitmask):
            for cid, serial in _iter_serial_enum_member(eid, value, bitmask):
                yield cid


def _iter_enum_ids():
    """Iterate the IDs of all enums in the IDB"""
    for index in range(idaapi.get_enum_qty()):
        yield idaapi.getn_enum(index)


def enums():
    """Iterate all enums in the IDB"""
    return (Enum(eid=eid) for eid in _iter_enum_ids())



import idaapi
import idautils
import idc

from . import base
from .. import core
from .. import exceptions

OPND_WRITE_FLAGS = {
    0: idaapi.CF_CHG1,
    1: idaapi.CF_CHG2,
    2: idaapi.CF_CHG3,
    3: idaapi.CF_CHG4,
    4: idaapi.CF_CHG5,
    5: idaapi.CF_CHG6,
}

OPND_READ_FLAGS = {
    0: idaapi.CF_USE1,
    1: idaapi.CF_USE2,
    2: idaapi.CF_USE3,
    3: idaapi.CF_USE4,
    4: idaapi.CF_USE5,
    5: idaapi.CF_USE6,
}


class Phrase(object):
    def __init__(self, op_t):
        self.op_t = op_t

        self._initialize()

    def _initialize(self):
        if self.op_t.type not in (idaapi.o_displ, idaapi.o_phrase):
            raise exceptions.OperandNotPhrase('Operand is not of type o_phrase or o_displ.')

        proc_name = idaapi.get_inf_structure().procname
        if proc_name != 'metapc':
            raise exceptions.PhraseProcessorNotSupported(
                'Phrase analysis not supported for processor {}'.format(proc_name))

        specflag1 = self.op_t.specflag1
        specflag2 = self.op_t.specflag2
        scale = 1 << ((specflag2 & 0xC0) >> 6)
        offset = self.op_t.addr

        if specflag1 == 0:
            index = None
            base_ = self.op_t.reg
        elif specflag1 == 1:
            index = (specflag2 & 0x38) >> 3
            base_ = (specflag2 & 0x07) >> 0

            if self.op_t.reg == 0xC:
                if base_ & 4:
                    base_ += 8
                if index & 4:
                    index += 8
        else:
            raise exceptions.PhraseNotSupported('o_displ, o_phrase : Not implemented yet : %x' % specflag1)

        # HACK: This is a really ugly hack. For some reason, phrases of the form `[esp + ...]` (`sp`, `rsp` as well)
        # set both the `index` and the `base` to `esp`. This is not significant, as `esp` cannot be used as an
        # index, but it does cause issues with the parsing.
        # This is only relevant to Intel architectures.
        if (index == base_ == idautils.procregs.sp.reg) and (scale == 1):
            index = None

        self.scale = scale
        self.index_id = index
        self.base_id = base_
        self.offset = offset

    @property
    def base(self):
        if self.base_id is None:
            return None
        return base.get_register_name(self.base_id)

    @property
    def index(self):
        if self.index_id is None:
            return None
        return base.get_register_name(self.index_id)

    def __repr__(self):
        phrase = []
        if self.base_id is not None:
            phrase.append(self.base)
        if self.index_id is not None:
            if phrase:
                phrase.append('+')
            phrase.append('{index}*{scale}'.format(index=self.index, scale=self.scale))
        if self.offset:
            offset = self.offset
            sign = '+'
            if core.is_signed(offset):
                offset = offset - (1 << (8 * core.get_native_size()))
                sign = '-'
            value = '{:X}'.format(abs(offset))
            phrase.append('{sign}{prefix}{value}{suffix}'.format(sign=sign if phrase or offset < 0 else '',
                                                                 prefix='0' if value[0].isalpha() else '',
                                                                 value=value,
                                                                 suffix='h' if abs(offset) > 9 else ''))

        return '[{}]'.format(''.join(phrase))


class OperandType(object):
    TYPES = {
        idaapi.o_void: "No_Operand",
        idaapi.o_reg: "General_Register",
        idaapi.o_mem: "Direct_Memory_Reference",
        idaapi.o_phrase: "Memory_Phrase",
        idaapi.o_displ: "Memory_Displacement",
        idaapi.o_imm: "Immediate_Value",
        idaapi.o_far: "Immediate_Far_Address",
        idaapi.o_near: "Immediate_Near_Address",
        idaapi.o_idpspec0: "Processor_specific_type",
        # There can be more processor specific types!
    }

    def __init__(self, type_):
        super(OperandType, self).__init__()

        self._type = type_

    @property
    def type(self):
        """Raw `type` value

        Use this if you need to pass the operand type around as a number.
        """
        return self._type

    @property
    def name(self):
        """Name of the xref type."""
        return self.TYPES.get(self._type, self.TYPES[idaapi.o_idpspec0])

    def __repr__(self):
        return self.name

    @property
    def is_void(self):
        return self._type == idaapi.o_void

    @property
    def is_reg(self):
        return self._type == idaapi.o_reg

    @property
    def is_mem(self):
        return self._type == idaapi.o_mem

    @property
    def is_phrase(self):
        return self._type == idaapi.o_phrase

    @property
    def is_displ(self):
        return self._type == idaapi.o_displ

    @property
    def is_imm(self):
        return self._type == idaapi.o_imm

    @property
    def is_far(self):
        return self._type == idaapi.o_far

    @property
    def is_near(self):
        return self._type == idaapi.o_near

    @property
    def is_special(self):
        return self._type >= idaapi.o_idpspec0

    @property
    def has_reg(self):
        return self._type in (idaapi.o_reg, idaapi.o_displ, idaapi.o_phrase)

    @property
    def has_phrase(self):
        return self._type in (idaapi.o_phrase, idaapi.o_displ)


class Operand(object):
    def __init__(self, operand, ea, insn, write=False, read=False):
        self._operand = operand
        self._write = write
        self._read = read
        self._type = OperandType(operand.type)
        self._ea = ea
        # We have to save the `insn_t` object referenced to make sure the `op_t` object is not released on the C side.
        self._insn = insn
        try:
            self._phrase = Phrase(operand)
        except exceptions.PhraseError:
            self._phrase = None

    @property
    def n(self):
        """Index of the operand in the instruction."""
        return self._operand.n

    @property
    def type(self):
        """Operand type."""
        return self._type

    @property
    def has_displacement(self):
        return base.operand_has_displacement(self._operand)

    @property
    def displacement(self):
        return base.operand_get_displacement(self._operand)

    offset = displacement

    @property
    def op_t(self):
        return self._operand

    @property
    def flags(self):
        return self._operand.flags

    @property
    def dtype(self):
        return self._operand.dtype

    @property
    def imm(self):
        return self._operand.value

    value = imm

    @property
    def addr(self):
        return self._operand.addr

    def has_reg(self, reg_name):
        return any(reg == reg_name for reg in self.regs)

    @property
    def size(self):
        """Size of the operand."""
        return base.dtype_to_size(self._operand.dtype)

    @property
    def is_read(self):
        """Is the operand value used in the instruction."""
        return self._read

    @property
    def is_write(self):
        """Is the operand value changed in the instruction."""
        return self._write

    @property
    def reg_id(self):
        """ID of the register used in the operand."""
        return self._operand.reg

    @property
    def reg(self):
        """Name of the register used in the operand."""
        if self.type.is_displ or self.type.is_phrase:
            size = core.get_native_size()
            return base.get_register_name(self.reg_id, size)

        if self.type.is_reg:
            return base.get_register_name(self.reg_id, self.size)

        else:
            raise exceptions.SarkOperandWithoutReg("Operand does not have a register.")

    @property
    def regs(self):
        if self.type.has_phrase:
            return set(reg for reg in (self.base, self.index) if reg)
        elif self.type.is_reg:
            return {base.get_register_name(self.reg_id, self.size)}
        else:
            return set()

    @property
    def text(self):
        return idc.print_operand(self._ea, self.n)

    def __str__(self):
        return self.text

    def __repr__(self):
        return "<Operand(n={}, text={!r})>".format(self.n, str(self))

    @property
    def base(self):
        if self._phrase:
            return self._phrase.base
        return self.reg

    @property
    def scale(self):
        if self._phrase:
            return self._phrase.scale
        return None

    @property
    def index(self):
        if self._phrase:
            return self._phrase.index
        return None

    @property
    def offset(self):
        if self._phrase:
            return self._phrase.offset
        return self.addr


class IndexingMode(object):
    def __init__(self, pre=False, post=False):
        self.pre = pre
        self.post = post

    @property
    def is_pre(self):
        return self.pre

    @property
    def is_post(self):
        return self.post

    @property
    def is_none(self):
        return not (self.pre or self.post)

    def __bool__(self):
        return self.pre or self.post


class Instruction(object):
    def __init__(self, ea):
        self._ea = ea
        self._insn = idautils.DecodeInstruction(ea)

        if self._insn is None:
            raise exceptions.SarkNoInstruction("No Instruction at 0x{:08X}.".format(ea))

        self._operands = self._make_operands()

    def __repr__(self):
        return f'<Instruction at 0x{self._ea:08x}>'

    def _make_operands(self):
        operands = []
        for index, operand in enumerate(self._insn.ops):
            if operand.type == idaapi.o_void:
                break  # No more operands.
            operands.append(Operand(operand,
                                    self._ea,
                                    insn=self._insn,
                                    write=self.is_operand_written_to(index),
                                    read=self.is_operand_read_from(index)))
        return operands

    @property
    def operands(self):
        """Instruction's Operands."""
        return self._operands

    @property
    def feature(self):
        """Canonical Features"""
        return self._insn.get_canon_feature()

    @property
    def mnem(self):
        """Instruction Mnemonic"""
        return self._insn.get_canon_mnem()

    def has_reg(self, reg_name):
        """Check if a register is used in the instruction."""
        return any(operand.has_reg(reg_name) for operand in self.operands)

    def is_operand_written_to(self, operand_index):
        """Check if an operand is written to (destination operand)."""
        return bool(self.feature & OPND_WRITE_FLAGS[operand_index])

    def is_operand_read_from(self, operand_index):
        """Check if an operand is read from (source operand)."""
        return bool(self.feature & OPND_READ_FLAGS[operand_index])

    @property
    def regs(self):
        """Names of all registers used by the instruction."""
        regs = set()
        for operand in self.operands:
            if not operand.type.has_reg:
                continue
            regs.update(operand.regs)
        return regs

    @property
    def is_call(self):
        """Is the instruction a call instruction."""
        return idaapi.is_call_insn(self._insn)

    @property
    def is_ret(self):
        """Is the instruction a return instruction."""
        return idaapi.is_ret_insn(self._insn)

    @property
    def is_indirect_jump(self):
        """Is the instruction an indirect jump instruction."""
        return idaapi.is_indirect_jump_insn(self._insn)

    @property
    def insn_t(self):
        return self._insn

    @property
    def indexing_mode(self):
        if idaapi.get_inf_structure().procname != 'ARM':
            return IndexingMode()

        return IndexingMode(pre=bool(self.insn_t.auxpref & 0x20),
                            post=bool(self.insn_t.auxpref & 0x80))



class SarkException(Exception):
    pass


class SarkError(SarkException):
    pass


class SarkNoSelection(SarkError):
    pass


class SarkNoFunction(SarkError):
    pass


class SarkAddFunctionFailed(SarkError):
    pass


class SarkFunctionExists(SarkError):
    pass


class SarkStructError(SarkError):
    pass


class SarkInvalidRegisterName(SarkError):
    pass


class SarkStructAlreadyExists(SarkStructError):
    pass


class SarkStructCreationFailed(SarkStructError):
    pass


class SarkStructNotFound(SarkStructError):
    pass


class SarkErrorAddStructMemeberFailed(SarkStructError):
    pass


class SarkErrorStructMemberName(SarkErrorAddStructMemeberFailed):
    pass


class SarkErrorStructMemberOffset(SarkErrorAddStructMemeberFailed):
    pass


class SarkErrorStructMemberSize(SarkErrorAddStructMemeberFailed):
    pass


class SarkErrorStructMemberTinfo(SarkErrorAddStructMemeberFailed):
    pass


class SarkErrorStructMemberStruct(SarkErrorAddStructMemeberFailed):
    pass


class SarkErrorStructMemberUnivar(SarkErrorAddStructMemeberFailed):
    pass


class SarkErrorStructMemberVarlast(SarkErrorAddStructMemeberFailed):
    pass


class SarkEnumError(SarkError):
    pass


class SarkErrorAddEnumMemeberFailed(SarkEnumError):
    pass


class SarkErrorEnumMemberName(SarkErrorAddEnumMemeberFailed):
    pass


class SarkErrorEnumMemberValue(SarkErrorAddEnumMemeberFailed):
    pass


class SarkErrorEnumMemberEnum(SarkErrorAddEnumMemeberFailed):
    pass


class SarkErrorEnumMemberMask(SarkErrorAddEnumMemeberFailed):
    pass


class SarkErrorEnumMemberIllv(SarkErrorAddEnumMemeberFailed):
    pass


class EnumNotFound(SarkEnumError):
    pass


class EnumCreationFailed(SarkEnumError):
    pass


class EnumAlreadyExists(SarkEnumError):
    pass


class CantRenameEnumMember(SarkEnumError):
    pass


class CantSetEnumMemberComment(SarkEnumError):
    pass


class SarkErrorNameAlreadyExists(SarkError):
    pass


class SarkSetNameFailed(SarkError):
    pass


class SarkSwitchError(SarkError):
    pass


class SarkNotASwitch(SarkSwitchError):
    pass


class SarkNoInstruction(SarkError):
    pass


class SarkOperandError(SarkError):
    pass


class SarkOperandWithoutReg(SarkOperandError):
    pass


class CantSetEnumComment(SarkEnumError):
    pass


class CantDeleteEnumMember(SarkEnumError):
    pass


class CantSetEnumBitfield(SarkEnumError):
    pass


class CantRenameEnum(SarkEnumError):
    pass


class SarkGuiError(SarkError):
    pass


class SarkMenuError(SarkGuiError):
    pass


class MenuAlreadyExists(SarkMenuError):
    pass


class MenuNotFound(SarkMenuError):
    pass


class FormNotFound(SarkGuiError):
    pass


class InvalidStructOffset(SarkStructError):
    pass


class SegmentError(SarkError):
    pass


class NoMoreSegments(SegmentError):
    pass


class InvalidBitness(SegmentError):
    pass


class NoFileOffset(SarkError):
    pass


class SarkNoString(SarkError):
    pass


class SarkExpectedPatchedByte(SarkError):
    pass


class PhraseError(SarkOperandError):
    pass


class OperandNotPhrase(PhraseError):
    pass


class PhraseNotSupported(PhraseError):
    pass


class PhraseProcessorNotSupported(PhraseNotSupported):
    pass


class SetTypeFailed(SarkError):
    def __init__(self, ea, c_signature):
        message = 'idc.SetType(ea={ea:08X}, "{c_signature}") failed'.format(ea=ea, c_signature=c_signature)
        super(SetTypeFailed, self).__init__(message)



import collections
import string

import idaapi
import idc
import idautils

from ..core import get_func, get_native_size
from .. import exceptions

NAME_VALID_CHARS = string.ascii_letters + string.digits + "?_:"

DTYPE_TO_SIZE = {
    idaapi.dt_byte: 1,
    idaapi.dt_word: 2,
    idaapi.dt_dword: 4,
    idaapi.dt_float: 4,
    idaapi.dt_double: 8,
    idaapi.dt_qword: 8,
    idaapi.dt_byte16: 16,
    idaapi.dt_fword: 6,
    idaapi.dt_byte32: 32,
    idaapi.dt_byte64: 64,
}


def dtype_to_size(dtyp):
    if dtyp == idaapi.dt_tbyte:
        # This can't be put in the dict as it depends on the current processor
        # and may change during runtime.
        return idaapi.ph_get_tbyte_size()

    return DTYPE_TO_SIZE[dtyp]


def is_ea_call(ea):
    inst = idautils.DecodeInstruction(ea)
    feature = inst.get_canon_feature()
    return feature & idaapi.CF_CALL


def get_register_info(reg_name):
    ri = idaapi.reg_info_t()
    if idaapi.IDA_SDK_VERSION >= 750:
        success = idaapi.parse_reg_name(ri, reg_name)
    else:
        success = idaapi.parse_reg_name(reg_name, ri)
    if not success:
        raise exceptions.SarkInvalidRegisterName("No register named {!r}".format(reg_name))
    return ri


def get_register_id(reg_name):
    return get_register_info(reg_name).reg


def get_register_size(reg_name):
    return get_register_info(reg_name).size


def get_register_name(reg_id, size=None):
    if size is None:
        size = get_native_size()
    return idaapi.get_reg_name(reg_id, size)


def operand_has_displacement(operand):
    if operand.type in (idaapi.o_phrase, idaapi.o_displ):
        return True

    return False


def operand_get_displacement(operand):
    return operand.addr


def is_same_function(ea1, ea2):
    try:
        if get_func(ea1).start_ea == get_func(ea2).start_ea:
            return True
    except:
        pass

    return False


Selection = collections.namedtuple("Selection", "start end")


def get_selection(always=True):
    start = idc.read_selection_start()
    end = idc.read_selection_end()

    if idaapi.BADADDR in (start, end):
        if not always:
            raise exceptions.SarkNoSelection()

        ea = idc.here()
        start = idaapi.get_item_head(ea)
        end = idaapi.get_item_end(ea)

    return Selection(start, end)


def get_highlighted_identifier():
    thing = idaapi.get_highlight(idaapi.get_current_viewer())
    if thing and thing[1]:
        return thing[0]

def format_name(name):
    try:
        return "".join(char if char in NAME_VALID_CHARS else "_" for char in name)
    except:
        return ""

def demangle(name, disable_mask=0):
    demangled_name = idaapi.demangle_name(name, disable_mask, idaapi.DQT_FULL)
    if demangled_name:
        return demangled_name
    return name


def get_offset_name(ea):
    # Try and get the function name
    try:
        func = get_func(ea)
        name = idaapi.get_ea_name(func.start_ea)
        name = demangle(name, 0x60) # MNG_NOTYPE | MNG_NORETTYPE
        if name:
            offset = ea - func.start_ea
            if offset:
                return '{}+{:X}'.format(name, offset)
            return name
    except exceptions.SarkNoFunction:
        pass

    # If that failed, use the segment name instead.
    segment = idaapi.getseg(ea)
    name = idaapi.get_segm_name(segment)
    offset_format = '{{:0{}X}}'.format(get_native_size() * 2)
    ea_text = offset_format.format(ea)
    if name:
        return '{}:{}'.format(name, ea_text)

    # Nothing found, simply return the address
    return ea_text



import os
import sys

from PyQt5 import QtCore, QtWidgets, QtGui

import idaapi

from . import exceptions


def capture_widget(widget, path=None):
    """Grab an image of a Qt widget

    Args:
        widget: The Qt Widget to capture
        path (optional): The path to save to. If not provided - will return image data.

    Returns:
        If a path is provided, the image will be saved to it.
        If not, the PNG buffer will be returned.
    """
    pixmap = widget.grab()

    if path:
        pixmap.save(path)

    else:
        image_buffer = QtCore.QBuffer()
        image_buffer.open(QtCore.QIODevice.ReadWrite)

        pixmap.save(image_buffer, "PNG")

        return image_buffer.data().data()


def get_widget(title):
    """Get the Qt widget of the IDA window with the given title."""
    tform = idaapi.find_widget(title)
    if not tform:
        raise exceptions.FormNotFound("No form titled {!r} found.".format(title))

    return idaapi.PluginForm.FormToPyQtWidget(tform)


def resize_widget(widget, width, height):
    """Resize a Qt widget."""
    widget.setGeometry(0, 0, width, height)


def get_window():
    """Get IDA's top level window."""
    tform = idaapi.get_current_widget()

    # Required sometimes when closing IDBs and not IDA.
    if not tform:
        tform = idaapi.find_widget("Output window")

    widget = idaapi.PluginForm.FormToPyQtWidget(tform)
    window = widget.window()
    return window


class MenuManager(object):
    """IDA Menu Manipulation

    Use this class to add your own top-level menus.
    While this is discouraged by the SDK:

    > You should not change top level menu, or the Edit,Plugins submenus

    (documentation for `attach_action_to_menu`, kernwin.hpp)

    Adding top-level menus is useful sometimes.
    Nonetheless, you should be careful and make sure to remove all your menus
    when you are done. Leaving them handing would force users to restart IDA
    to remove them.

    Usage of this class should be as follows:

    >>> # Use the manager to add top-level menus
    >>> menu_manager = MenuManager()
    >>> menu_manager.add_menu("My Menu")
    >>> # Use the standard API to add menu items
    >>> idaapi.attach_action_to_menu("My Menu/", ":My-Action:", idaapi.SETMENU_APP)
    >>> # When a menu is not needed, remove it
    >>> menu_manager.remove_menu("My Menu")
    >>> # When you are done with the manager (and want to remove all menus you added,)
    >>> # clear it before deleting.
    >>> menu_manager.clear()
    """

    def __init__(self):
        super(MenuManager, self).__init__()

        self._window = get_window()
        self._menu = self._window.findChild(QtWidgets.QMenuBar)

        self._menus = {}

    def add_menu(self, name):
        """Add a top-level menu.

        The menu manager only allows one menu of the same name. However, it does
        not make sure that there are no pre-existing menus of that name.
        """
        if name in self._menus:
            raise exceptions.MenuAlreadyExists("Menu name {!r} already exists.".format(name))
        menu = self._menu.addMenu(name)
        self._menus[name] = menu

    def remove_menu(self, name):
        """Remove a top-level menu.

        Only removes menus created by the same menu manager.
        """
        if name not in self._menus:
            raise exceptions.MenuNotFound(
                "Menu {!r} was not found. It might be deleted, or belong to another menu manager.".format(name))

        self._menu.removeAction(self._menus[name].menuAction())
        del self._menus[name]

    def clear(self):
        """Clear all menus created by this manager."""
        for menu in self._menus.values():
            self._menu.removeAction(menu.menuAction())
        self._menus = {}



from collections import namedtuple
import idc
import idaapi

import shutil
import itertools
import struct
from .core import fix_addresses, get_native_size
from . import exceptions


def Bytes(start=None, end=None):
    start, end = fix_addresses(start, end)

    return map(idaapi.get_wide_byte, list(range(start, end)))


def Words(start=None, end=None):
    start, end = fix_addresses(start, end)

    return map(idaapi.get_wide_word, list(range(start, end, 2)))


def Dwords(start=None, end=None):
    start, end = fix_addresses(start, end)

    return map(idaapi.get_wide_dword, list(range(start, end, 4)))


def Qwords(start=None, end=None):
    start, end = fix_addresses(start, end)

    return map(idaapi.get_qword, list(range(start, end, 8)))


def NativeWords(start=None, end=None):
    native_size = get_native_size()

    if native_size == 2:
        return Words(start, end)
    elif native_size == 4:
        return Dwords(start, end)
    elif native_size == 8:
        return Qwords(start, end)


def bytes_until(byte=0, start=None, end=None):
    return iter(Bytes(start, end).__next__, byte)


def words_until(word=0, start=None, end=None):
    return iter(Words(start, end).__next__, word)


def dwords_until(dword=0, start=None, end=None):
    return iter(Dwords(start, end).__next__, dword)


def qwords_until(qword=0, start=None, end=None):
    return iter(Qwords(start, end).__next__, qword)


def native_words_until(native_word=0, start=None, end=None):
    return iter(NativeWords(start, end).__next__, native_word)


def Chars(start=None, end=None):
    return map(chr, Bytes(start, end))


def chars_until(char='\0', start=None, end=None):
    return iter(Chars(start, end).__next__, char)


def read_ascii_string(ea, max_length=None):
    if max_length is None:
        end = None
    else:
        end = ea + max_length
    return "".join(chars_until(start=ea, end=end))


def dword_to_bytes(dword):
    return struct.pack(">L", dword)


def read_memory(start, end):
    size = end - start
    return idaapi.get_bytes(start, size)


def write_memory(start, data, destructive=False):
    if destructive:
        idaapi.put_bytes(start, data)

    else:
        idaapi.patch_bytes(start, data)


PatchedByte = namedtuple("PatchedByte", "ea fpos original patched")


def get_patched_bytes(start=None, end=None):
    start, end = fix_addresses(start, end)

    patched_bytes = dict()

    def collector(ea, fpos, original, patched):
        patched_bytes[ea] = PatchedByte(ea, fpos, original, patched)
        return 0

    idaapi.visit_patched_bytes(start, end, collector)

    return patched_bytes


def apply_patches(output_path=None):
    to_patch = idaapi.get_input_file_path()

    if output_path:
        shutil.copyfile(to_patch, output_path)
        to_patch = output_path

    patches = get_patched_bytes()

    with open(to_patch, "r+b") as output:
        for patch in patches.values():
            output.seek(patch.fpos)
            patched_byte = bytes([patch.patched])
            output.write(patched_byte)


def undefine(start, end):
    idaapi.del_items(start, idaapi.DELIT_SIMPLE, end - start)


def is_string(ea):
    string_type = idc.get_str_type(idaapi.get_item_head(ea))

    if string_type in (None, -1):
        return False

    return True


def get_string(ea):
    """Read the string at the given ea.

    This function uses IDA's string APIs and does not implement any special logic.
    """
    # We get the item-head because the `GetStringType` function only works on the head of an item.
    string_type = idc.get_str_type(idaapi.get_item_head(ea))

    if string_type is None:
        raise exceptions.SarkNoString("No string at 0x{:08X}".format(ea))

    string = idc.get_strlit_contents(ea, strtype=string_type)

    if not string:
        raise exceptions.SarkNoString("No string at 0x{:08X}".format(ea))

    return string



from collections import namedtuple, defaultdict
import operator

import idaapi
import idautils
import idc

from . import exceptions
from .code import lines

FF_TYPES = [idc.FF_BYTE, idc.FF_WORD, idc.FF_DWORD, idc.FF_QWORD, idc.FF_OWORD, ]
FF_SIZES = [1, 2, 4, 8, 16, ]

SIZE_TO_TYPE = dict(zip(FF_SIZES, FF_TYPES))

STRUCT_ERROR_MAP = {
    idc.STRUC_ERROR_MEMBER_NAME:
        (exceptions.SarkErrorStructMemberName, "already has member with this name (bad name)"),
    idc.STRUC_ERROR_MEMBER_OFFSET:
        (exceptions.SarkErrorStructMemberOffset, "already has member at this offset"),
    idc.STRUC_ERROR_MEMBER_SIZE:
        (exceptions.SarkErrorStructMemberSize, "bad number of bytes or bad sizeof(type)"),
    idc.STRUC_ERROR_MEMBER_TINFO:
        (exceptions.SarkErrorStructMemberTinfo, "bad typeid parameter"),
    idc.STRUC_ERROR_MEMBER_STRUCT:
        (exceptions.SarkErrorStructMemberStruct, "bad struct id (the 1st argument)"),
    idc.STRUC_ERROR_MEMBER_UNIVAR:
        (exceptions.SarkErrorStructMemberUnivar, "unions can't have variable sized members"),
    idc.STRUC_ERROR_MEMBER_VARLAST:
        (exceptions.SarkErrorStructMemberVarlast, "variable sized member should be the last member in the structure"),
}


def struct_member_error(err, sid, name, offset, size):
    """Create and format a struct member exception.

    Args:
        err: The error value returned from struct member creation
        sid: The struct id
        name: The member name
        offset: Memeber offset
        size: Member size

    Returns:
        A ``SarkErrorAddStructMemeberFailed`` derivative exception, with an
        informative message.
    """
    exception, msg = STRUCT_ERROR_MAP[err]
    struct_name = idaapi.get_struc_name(sid)
    return exception(('AddStructMember(struct="{}", member="{}", offset={}, size={}) '
                      'failed: {}').format(
        struct_name,
        name,
        offset,
        size,
        msg
    ))


def create_struct(name):
    """Create a structure.

    Args:
        name: The structure's name

    Returns:
        The sturct ID

    Raises:
        exceptions.SarkStructAlreadyExists: A struct with the same name already exists
        exceptions.SarkCreationFailed:  Struct creation failed
    """
    sid = idaapi.get_struc_id(name)
    if sid != idaapi.BADADDR:
        # The struct already exists.
        raise exceptions.SarkStructAlreadyExists("A struct names {!r} already exists.".format(name))

    sid = idaapi.add_struc(idaapi.BADADDR, name, 0)
    if sid == idaapi.BADADDR:
        raise exceptions.SarkStructCreationFailed("Struct creation failed.")

    return sid


def get_struct(name):
    """Get a struct by it's name.

    Args:
        name: The name of the struct

    Returns:
        The struct's id

    Raises:
        exceptions.SarkStructNotFound: is the struct does not exist.
    """
    sid = idaapi.get_struc_id(name)
    if sid == idaapi.BADADDR:
        raise exceptions.SarkStructNotFound()

    return sid


def size_to_flags(size):
    return SIZE_TO_TYPE[size] | idc.FF_DATA


def add_struct_member(sid, name, offset, size):
    failure = idc.add_struc_member(sid, name, offset, size_to_flags(size), -1, size)

    if failure:
        raise struct_member_error(failure, sid, name, offset, size)


StructOffset = namedtuple("StructOffset", "offset size")
OperandRef = namedtuple("OperandRef", "ea n")


def infer_struct_offsets(start, end, reg_name):
    offsets = set()
    operands = []
    for line in lines(start, end):
        for operand in line.insn.operands:
            if not operand.has_reg(reg_name):
                continue

            if not operand.type.has_phrase:
                continue

            if not operand.base:
                continue

            offset = operand.offset
            if offset < 0:
                raise exceptions.InvalidStructOffset(
                    "Invalid structure offset 0x{:08X}, probably negative number.".format(offset))
            size = operand.size
            offsets.add(StructOffset(offset, size))
            operands.append(OperandRef(line.ea, operand.n))

    return offsets, operands


def get_common_register(start, end):
    """Get the register most commonly used in accessing structs.

    Access to is considered for every opcode that accesses memory
    in an offset from a register::

        mov eax, [ebx + 5]

    For every access, the struct-referencing registers, in this case
    `ebx`, are counted. The most used one is returned.

    Args:
        start: The adderss to start at
        end: The address to finish at
    """
    registers = defaultdict(int)
    for line in lines(start, end):
        insn = line.insn

        for operand in insn.operands:

            if not operand.type.has_phrase:
                continue

            if not operand.base:
                continue

            register_name = operand.base
            registers[register_name] += 1

    return max(registers.items(), key=operator.itemgetter(1))[0]


def offset_name(offset):
    """Format an offset into a name."""
    return "offset_{:X}".format(offset.offset)


def set_struct_offsets(offsets, sid):
    for offset in offsets:
        try:
            add_struct_member(sid,
                              offset_name(offset),
                              offset.offset,
                              offset.size)
        except exceptions.SarkErrorStructMemberName:
            # Get the offset of the member with the same name
            existing_offset = idc.get_member_offset(sid, offset_name(offset))
            if offset.offset == existing_offset:
                pass
            else:
                raise
        except exceptions.SarkErrorStructMemberOffset:
            # Get the size of the member at the same offset
            if offset.size == idc.get_member_size(sid, offset.offset):
                # If they are the same, all is well.
                pass


def create_struct_from_offsets(name, offsets):
    sid = create_struct(name)

    set_struct_offsets(offsets, sid)


def apply_struct(start, end, reg_name, struct_name):
    offsets, operands = infer_struct_offsets(start, end, reg_name)

    sid = get_struct(struct_name)

    for ea, n in operands:
        insn = idautils.DecodeInstruction(ea)
        idc.op_stroff(insn, n, sid, 0)


def selection_has_offsets(start, end):
    for line in lines(start, end):
        for operand in line.insn.operands:
            if not operand.type.has_phrase:
                continue
            if not operand.base:
                continue
            return True
    return False
